/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Dynamic extensions loading.
 * 
 *  This file provides the logic to load a dynamic extension after the daemon has initialized 
 * all its internal states.
 *  During the configuration parsing, each extension that must be loaded is registered with
 * ext_register(extension_filename, conf_filename);
 *  Then when the parsing is terminated, the extensions are loaded in turn with
 * ext_load();
 *
 *  ext_init and ext_fini are called at daemon init and termination.
 * 
 * The normal call flow of this module is:
 *    ext_init
 *    ext_register
 *    ext_register
 *    ext_register
 *    ext_register
 *    ext_load
 *    (no more interaction until daemon shutdown)
 *    ext_fini
 */

#ifndef _EXTENSIONS_H
#define _EXTENSIONS_H

/*
 * FUNCTION:	ext_init
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *   Initialize the extension module. Must be called before any call to another function of the module.
 *
 * RETURN VALUE:
 *  0      	: Module is initialized.
 *  !0 		: An error occured (standard errno are returned).
 */
int ext_init( void );

/*
 * FUNCTION:	ext_fini
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *   On daemon shutdown, an attempt is made to unload all loaded extensions.
 *
 * RETURN VALUE:
 *  0      	: Module is cleaned.
 *  !0 		: An error occured (standard errno are returned). The error may usually be ignored.
 */
int ext_fini( void );

/*
 * FUNCTION:	ext_register
 *
 * PARAMETERS:
 *  filename 	: the filename of the extension. The filename argument is used to construct a pathname to the object file. 
 *		 If filename contains a slash character, the filename argument is used as the pathname for the file. 
 *		 Otherwise, filename is used in an implementation-defined manner to yield a pathname.
 *  conffile 	: the filename of the extension configuration file. NULL is passed if no conffile is given in waaad configuration file.
 *
 * DESCRIPTION: 
 *   Save the filename of the extension and its configuration file, for loading it later when ext_load is called. 
 *  The extensions are loaded in FIFO order of their registration.
 *
 * RETURN VALUE:
 *  0      	: Filename is registered for loading later.
 *  EINVAL 	: Invalid parameter.
 *  EBUSY 	: ext_load has already been called.
 *  ENOMEM 	: Memory allocation for the new object element failed.
 */
int ext_register( char * filename, char * conffile );

/*
 * FUNCTION:	ext_load
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *  Load extensions that have been registered by ext_register. 
 *  All extensions must provide a waaad_ext_init entry point, that is called.
 *
 * RETURN VALUE:
 *  0      	: All extensions are loaded successfully.
 *  EINVAL 	: At list one extension has failed to load.
 */
int ext_load(void);

#endif /* ! _EXTENSIONS_H */
