/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the mechanism for managing dynamic peers expiry.
 *
 */
#include "waaad-internal.h"
#include "peer-internal.h"

/* Mutex to protect the lists of peers */
pthread_mutex_t _peers_lists_mtx = PTHREAD_MUTEX_INITIALIZER;

/* List of deleted peers in grace period */
uti_list_t 	_peers_deleted;
/* List of dynamic peers that expires with a timer */
uti_list_t 	_peers_expiring;

/* Condvar to use in the expiry thread */
static pthread_cond_t	_peers_lists_cond = PTHREAD_COND_INITIALIZER;

/* The expiry thread */
static pthread_t expire_thread;


/* This function is called when a peer must be queued in the expiry list (creation or update) */
int _peer_expire_queue(_peer_t * peer)
{
	uti_list_t * li;
	int first = 1;
	
	CHECK_PARAMS(  peer 
			&& IS_LIST_EMPTY(&peer->p_exp_list) 
			&& (peer->p_addinfo.pa_flags & PA_EXPIRY) 
			&& (peer->p_addinfo.pa_ts.tv_sec != 0)  );
	
	/* lock the queue */
	CHECK_POSIX(  pthread_mutex_lock(&_peers_lists_mtx)  );
	
	/* Now find the place for this thread */
	for (li = _peers_expiring.next; li != &_peers_expiring; li = li->next) {
		_peer_t * cur = _P(li->o);
		
		if (TS_IS_INFERIOR( &(peer->p_addinfo.pa_ts), &(cur->p_addinfo.pa_ts) ))
			break;
		
		first = 0;
	}
	
	/* Now insert just before this place */
	uti_list_insert_before(li, &peer->p_exp_list);
	
	/* In case we inserted in the first position, signal the condvar */
	if (first) {
		CHECK_POSIX(  pthread_cond_signal(&_peers_lists_cond)  );
	}
	
	/* unlock the queue */
	CHECK_POSIX(  pthread_mutex_unlock(&_peers_lists_mtx)  );
	
	return 0;
}

/* The code of the thread to send an EXPIRE event / delete the peer when the timeout timer is reached. */
static void * _expire_th (void * arg)
{
	THREAD_NAME( "Expiry" );
			
	TRACE_ENTRY( "%p", arg );
	
	/* Lock the lists lock. It will be released on cancellation */
	CHECK_POSIX_DO(  pthread_mutex_lock( &_peers_lists_mtx ),  goto error  );
	pthread_cleanup_push( cleanup_mutex, &_peers_lists_mtx);
	
	/* loop forever, the thread will be canceled */
	while (1) {
		struct timespec	now;
		_peer_t * nextpeer = NULL;
	
		/* Pick the next timer, either head of g_peers_expiring or g_peers_deleted */
		if (!IS_LIST_EMPTY(&_peers_expiring)) {
			nextpeer = _P(_peers_expiring.next->o);
		}
		
		if (!IS_LIST_EMPTY(&_peers_deleted)) {
			_peer_t * t = _P(_peers_deleted.next->o);
			
			if ( TS_IS_INFERIOR(&t->p_addinfo.pa_ts, &nextpeer->p_addinfo.pa_ts) ) {
				nextpeer = t;
			}
		}
	
		/* if no thread, just wait for the condvar to be signaled, then loop. */
		if (nextpeer == NULL) {
			CHECK_POSIX_DO(  pthread_cond_wait( &_peers_lists_cond, &_peers_lists_mtx ), goto error  );
			continue;
		}
		
		/* Get the current time */
		CHECK_SYS_DO(  clock_gettime(CLOCK_REALTIME, &now),  goto error  );

		/* if timer is in the future, we only wait until that time comes */
		if ( TS_IS_INFERIOR( &now, &nextpeer->p_addinfo.pa_ts ) ) {
			
			CHECK_POSIX_DO2(  pthread_cond_timedwait( &_peers_lists_cond, &_peers_lists_mtx, &nextpeer->p_addinfo.pa_ts ),  
					ETIMEDOUT, /* ETIMEDOUT is a normal error, continue */,
					/* on other error, */ goto error );
	
			/* on wakeup, loop */
			continue;
		}
	
		/* the timer is expired*/
		
		/* For peer that just expired, we send the PEVENT_EXPIRE event to the PSM. 
		  It is responsible for refreshing and requeuing or grace deleting. */
		if (nextpeer->p_exp_list.head == &_peers_expiring) {
			uti_list_unlink(&nextpeer->p_exp_list);
			/* for expiring thread, remove from this list and send a PEVENT_EXPIRE event */
			CHECK_FCT_DO(  _peer_events_send(nextpeer, PEVENT_EXPIRE, NULL),  goto error  );
		} else {
			/* for deleted threads, really delete the thread */
			uti_list_unlink(&nextpeer->p_exp_list);
			
			/* Remove from the global hash */
			CHECK_FCT_DO(  _peer_struct_unlink(nextpeer),  goto error  );
			
			/* And free it */
			CHECK_FCT_DO(  _peer_struct_destroy(&nextpeer),  goto error  );
		}
		
		/* and loop */
	}
	
error:	
	/* We arrive here only in case of error => we should stop the daemon */
	CHECK_FCT_DO(  uti_event_send(WDE_ERROR),  /* nothing */  );

	pthread_cleanup_pop(0);

	return NULL;
}


/* Initialize the mechanism */
int _peer_expire_init(void)
{
	uti_list_init(&_peers_expiring, NULL);
	uti_list_init(&_peers_deleted, NULL);
	
	/* Create the expire thread */
	CHECK_POSIX(  pthread_create(&expire_thread, NULL, _expire_th, NULL)  );
	
	return 0;
}

/* Terminate the mechanism */
int _peer_expire_fini(void)
{
	/* Kill the thread */
	CHECK_FCT_DO( _thread_term(&expire_thread), /* continue */ );
	
	/* The peers will be actually deleted when the hash table is destructed */
	return 0;
}


