/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the functions to initialize and terminate the module.
 *
 */
#include "waaad-internal.h"
#include "peer-internal.h"

/* Initialize the module */
int peer_init ( void )
{
	TRACE_ENTRY( "" );
	
	/* Initialize the hash table */
	CHECK_FCT( _peer_struct_init()  );
	
	/* Initialize the expiry mechanism */
	CHECK_FCT( _peer_expire_init()  );
	
	/* Initialize the dictionary objects */
	CHECK_FCT( _peer_cer_cea_init() );
	CHECK_FCT( _peer_dwr_dwa_init() );
	CHECK_FCT( _peer_dpr_dpa_init() );
	
	/* Note: the server and psms are created later, when peer_start is called */

	/* Done! */
	return 0;
}

/* End of the module */
int peer_fini ( void )
{
	TRACE_ENTRY( "" );
	
	/* Stop the server to not accept new connections */
	CHECK_FCT_DO(  _peer_server_stop(), /* continue */  );
	
	/* Destroy all the peers and the hash table */
	CHECK_FCT_DO(  _peer_struct_fini(), /* continue */  );
	
	/* Stop the expiry thread. */
	CHECK_FCT_DO(  _peer_expire_fini(), /* continue */  );
	
	/* Destroy the messages helpers */
	CHECK_FCT_DO(  _peer_cer_cea_fini(), /* continue */  );
	CHECK_FCT_DO(  _peer_dwr_dwa_fini(), /* continue */  );
	
	/* We're done */
	return 0;
}

/* Has peer_start() been called? */
static int _peer_started = 0;

/* Start the listening sockets and the peers state machines */
int peer_start ( void )
{
	if (_peer_started)
		return EINVAL;
	
	/* Start the peers PSMs */
	CHECK_FCT(  _peer_psm_start_all()  );
	
	/* Accept incoming connections */
	CHECK_FCT(  _peer_server_start()  );
	
	_peer_started = 1;
	
	return 0;
}
	
/* Get some information from a peer_t opaque object */
int peer_get ( peer_t * peer, peer_info_t what, void **result )
{
	TRACE_ENTRY( "%p %d %p", peer, what, result );
	
	/* Verify the parameters */
	CHECK_PARAMS(  VALIDATE_PEER( peer ) && result  );
	
	/* now return the value */
	switch (what) {
		case PEER_DIAMETER_ID:
			*result = _P(peer)->p_diamid;
			break;
		
		case PEER_REALM:
			*result = _P(peer)->p_realm;
			break;
		
		case PEER_ONLINE:
			*result = (void *)(unsigned long)(_P(peer)->p_state == STATE_OPEN);
			break;
		
		case PEER_APPLICATIONS:
			*result = _P(peer)->p_app_list;
			break;
		
		default:
			CHECK_PARAMS( 0 ); /* log error and return EINVAL */
	}
	
	return 0;
}


/* Add a new peer */
int peer_add ( char * diamid, size_t len, peer_add_t * data )
{
	int ret = 0;
	int i;
	uint32_t hash;
	_peer_t * new;
	uti_list_t *next;
	char * di;
	char * rlm;
	
	TRACE_ENTRY( "%p %d %p", diamid, len, data );
	
	/* Check the parameters */
	CHECK_PARAMS(  diamid && len && data  );

	/* Check the data */
#ifdef DISABLE_SCTP
	CHECK_PARAMS(  (data->pa_proto == IPPROTO_TCP)  );
#else /* DISABLE_SCTP */
	CHECK_PARAMS(  (data->pa_proto == IPPROTO_SCTP) || (data->pa_proto == IPPROTO_TCP)  );
#endif /* DISABLE_SCTP */
	
	/* Create the diameter id string that will be saved */
	CHECK_MALLOC(  di = (char *)malloc(len + 1)  );
	
	for (i=0; i<len; i++) {
		di[i] = (char)tolower(diamid[i]);
	}
	di[len] = '\0';
	
	CHECK_PARAMS( rlm = strchr(di, '.') );
	rlm++; /* we suppose the realm starts after the first '.' */
	
	/* Check we don't have this peer already. This locks the hash table and returns the next list item in 'next' */
	/* It is not very efficient to lock the hash line for a long time, but adding a peer is quite rare so it is OK */
	ret = _peer_struct_search(di, len, &hash, &next);
	if (ret == 0) {
		TRACE_DEBUG(INFO, "A peer already exists with this diameter id ('%s')", di);
		CHECK_POSIX(  pthread_mutex_unlock(H_LOCK(hash))  );
		ret = EALREADY;
		goto error;
	}
	if (ret != ENOENT) {
		/* Another error occurred */
		CHECK_POSIX(  pthread_mutex_unlock(H_LOCK(hash))  );
		CHECK_FCT(ret);
	}
	
	/* create a new object */
	CHECK_FCT_DO(  _peer_struct_malloc(di, &new), { ret = ENOMEM; goto error; }  );
	
	/* link it into the hash table */
	uti_list_insert_before(next, &new->p_global);
	H_COUNT(hash)++;
	
	CHECK_POSIX(  pthread_mutex_unlock(H_LOCK(hash))  );
	
	/* Save additional data into the peer */
	new->p_hash = hash;
	
	/* Note about the realm, we may want to pass it as parameter later */
	CHECK_MALLOC_DO(  new->p_realm = strdup(rlm),   { ret = ENOMEM; goto error; }  );
	
	memcpy(&new->p_addinfo, data, sizeof(peer_add_t));
	
	/* If peer expires, link also into the expiry list */
	if ( (data->pa_flags & PA_EXPIRY) && (data->pa_ts.tv_sec != 0) ) {
		CHECK_FCT(  _peer_expire_queue(new)  );
	}
	
	/* if _peer_started, create the psm thread */
	if (_peer_started) {
		CHECK_FCT(  _peer_psm_start(new, 1) );
	}
	
	return 0;

/* rollback in case of error : */
error:	
	if (new) {
		_peer_struct_unlink(new);
		_peer_struct_destroy(&new);
	}
	return ret;
}


/* Queue a message in the given peer. The end-to-end id must already be set. The hbh will be set automatically for requests. */
int peer_send ( peer_t * peer, msg_t * msg )
{
	int ret = 0;
	
	TRACE_ENTRY( "%p %p", peer, msg );
	
	/* Verify the parameters */
	CHECK_PARAMS(  VALIDATE_PEER( peer )  &&  msg  );
	
	/* Lock the peer object */
	CHECK_POSIX(  pthread_mutex_lock( &(_P(peer)->p_lock) )  );
	
	/* Check the state of the peer is valid */
	if (_P(peer)->p_state != STATE_OPEN) {
		TRACE_DEBUG(INFO, "Peer '%s' is not in OPEN state, do not accept message", _P(peer)->p_diamid);
		CHECK_POSIX(  pthread_mutex_unlock( &(_P(peer)->p_lock) )  );
		return ENOTCONN;
	}
	
	/* put the message in the queue */
	CHECK_FCT_DO(  ret = meq_post( _P(peer)->p_out_q, msg ),  
		{
			CHECK_POSIX(  pthread_mutex_unlock( &(_P(peer)->p_lock) )  );
			return ret;
		}  );
	
	/* unlock and return */
	CHECK_POSIX(  pthread_mutex_unlock( &(_P(peer)->p_lock) )  );
	
	return 0;
}

